#!/bin/sh
#
# Copyright (c) 2014, The Linux Foundation. All rights reserved.
# Copyright (c) 2016, GuoGuo<gch981213@gmail.com>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
. /lib/netifd/netifd-wireless.sh
. /lib/netifd/hostapd.sh
. /lib/functions.sh

init_wireless_driver "$@"

# DEBUG=1
wlanconfig() {
	[ -n "${DEBUG}" ] && echo wlanconfig "$@"
	/usr/sbin/wlanconfig "$@"
}

iwconfig() {
	[ -n "${DEBUG}" ] && echo iwconfig "$@"
	/usr/sbin/iwconfig "$@"
}

iwpriv() {
	[ -n "${DEBUG}" ] && echo iwpriv "$@"
	/usr/sbin/iwpriv "$@"
}

drv_qcawifi_init_device_config() {
	config_add_string acs_chloadvar acs_ctrlflags acs_dbgtrace acs_lmtobss acs_rssivar acs_scanintvl acs_txpwr_opt AMPDU \
		ampdudensity AMPDUFrames AMPDULim AMPDURxBsize antenna ant_numitr ant_numpkts ant_pktlen ant_retrain ant_traffic_timer \
		ant_train ant_train_min_thres ant_trainmode ant_train_thres ant_traintype burst_dur channel country current_ant dcs_coch_int \
		dcs_debug dcs_enable dcs_errth dcs_phyerrth dcs_usermaxc default_ant distance dscp_tid_map eppovrd_ch_144 hmmc_dscp_tid_map \
		htmode hwmode igmp_dscp_tid_map macaddr mcast_echo obss_long_slot obss_rssi_th obss_rx_rssi_th phy promisc qwrap_enable \
		reset_dscp_map retrain_drop retrain_interval rxantenna rxchainmask set_ch_144 set_pmf set_sa_param set_smart_antenna \
		staDFSEnable txantenna txbf_snd_int txchainmask txpower TXPowLim2G TXPowLim5G
	config_add_array aggr_burst
	config_add_boolean acs_bkscanen aldstats allowpromisc AMSDU ani_enable ap_isolation_enabled bcnburst blk_report_fld block_interbss \
		burst diversity drop_sta_query dscp_ovride enable_ol_stats hmmc_dscp_ovride igmp_dscp_ovride nosbeacon rst_lro_stats rst_sg_stats \
		rst_tso_stats set_fw_recovery softled wps_pbc_overwrite_ap_settings wps_pbc_skip_ap_if_sta_disconnected wps_pbc_try_sta_always noscan
}

drv_qcawifi_init_iface_config() {
	hostapd_common_add_bss_config
	config_add_string acparams acsmaxdwell acsmindwell acsreport addba addbaresp aifs ampdu amsdu athnewind bintval cca_thresh \
		cfreq2 chanbw chbwmode ch_cnt_th ch_cntwn_dur ch_hop_en ch_long_dur ch_nhop_dur ch_noise_th chwidth clr_rawsim_stats cwmax \
		cwmin dbgLVL decap_type delba do_tdls_dc_req doth_chanswitch dtim_period enablertscts encap_type encryption extap \
		extprotmode extprotspac frag ifname inact ldpc maxampdu maxsta mcast_rate me_adddeny metimeout \
		metimer mfptest mode nawds_defcaps nawds_mode nawds_override noackpolicy noedgech noIBSSCreate nss offchan_tx_test osen periodicScan \
		powersave ps_on_time ps_timeout puren quiet rrmdbg rts rxcorrection rxfilterinvalid rx_stbc scanband set11NRates set11NRetries \
		setaddbaoper setibssdfsparam setibssrssiclass setibssrssihyst set_monrxfilter set_tdls_rmac setwmmparams sko_max_xretries \
		startibssrssimon tdls tdlsaction tdls_auto tdls_dtoken tdlsecchnoffst tdlsmacaddr1 tdlsmacaddr2 tdls_margin tdlsoffchan tdlsoffchnmode \
		tdls_off_timeout tdls_path_sel tdls_path_sel_period tdls_qosnull tdls_rssi_lb tdls_rssi_offset tdls_rssi_ub tdls_set_rcpi tdls_set_rcpi_hi \
		tdls_set_rcpi_lo tdls_set_rcpi_margin tdlsswitchtime tdls_tdb_timeout tdlstimeout tdls_uapsd tdls_weak_timeout TxBFCTL txcorrection \
		txoplimit txpower txqueuelen tx_stbc vap_ind vhtmaxampdu vhtmcs vht_mcsmap vhtsounddim vhtstscap wnm wps_config wps_pbc_duration \
		wps_pbc_start_time
	config_add_boolean ant_ps_on blockdfschan countryie cwmenable dgaf_disable noscan doth ext_ifu_acs implicitbf isolate \
	l2tif mcastenhance medropmcast protmode proxyarp qbssload rawsim_debug rawsim_txagr rrm rrmslwin rrmsstats scanchevent send_add_ies \
	shortgi stafwd sw_merge uapsd vht_11ng vhtmubfee vhtmubfer vhtsubfee vhtsubfer vow_dbg wmm wps_pbc wps_pbc_enable
	config_add_array vow_dbg_cfg
	config_add_array set_max_rate
}

check_qcawifi_phy() {
	local phy="$1"
	[ -n "$phy" -a -d "/sys/class/net/$phy" ] || {
		echo "qcawifi:phy for wifi device $1 not found"
		return 1
	}
	return 0
}

qcawifi_hostapd_setup_vif() {
	local ifname="$1" && shift
	local driver="atheros"
	echo "qcawifi:Start hostapd for $ifname"
	hostapd_cfg=
	hostapd_set_log_options hostapd_cfg "$device"
	hostapd_set_bss_options hostapd_cfg "$device" "$vif"
	case "$hwmode" in
		*bg|*gdt|*gst|*fh) hwmode=g;;
		*adt|*ast) hwmode=a;;
	esac
	cat > /var/run/hostapd-$ifname.conf <<EOF
driver=atheros
interface=$ifname
$hostapd_cfg
EOF
	entropy_file=/var/run/entropy-$ifname.bin

	[ -f "/var/run/hostapd-$ifname.lock" ] &&
		rm /var/run/hostapd-$ifname.lock
	wpa_cli -g /var/run/hostapd/global raw ADD bss_config=$ifname:/var/run/hostapd-$ifname.conf
	ret="$?"
	touch /var/run/hostapd-$ifname.lock

	[ "$ret" != 0 ] && {
		wireless_setup_failed HOSTAPD_START_FAILED
		return
	}
	return 0
}

qcawifi_wpa_supplicant_setup_vif() {
	local ifname=$1
	echo "Start supplicant for vif $ifname"
	wpa_supplicant_prepare_interface "$ifname" athr || return 1
	wpa_supplicant_add_network "$ifname"
	wpa_supplicant_run "$ifname"
}

apply_config_to_if() {
	local dest_if=$1;shift
	while [ -n "$1" ]
	do
		json_get_var cfg_tval $1
		[ -n "$cfg_tval" ] && iwpriv "$dest_if" $1 "${cfg_tval}"
		shift
	done
}

qcawifi_vif_setup() {
	local start_hostapd= vif_txpower= wlanaddr=""
	local wlanmode
	local name="$1"
	json_select config
	json_get_var ifname ifname "ath${radioidx}$ifidx"
	json_get_var enc encryption "none"
	json_get_var eap_type eap_type
	json_get_var mode mode
	let ifidx+=1
	wlanmode=$mode
	[ "$wlanmode" = "ap_monitor" ] && wlanmode="specialvap"
	[ "$wlanmode" = "ap_smart_monitor" ] && wlanmode="smart_monitor"
	echo "qcawifi: Set up $mode vif $ifname."
	case "$mode" in
		sta)
			[ $qwrap_enable -gt 0 ] && wlanaddr="00:00:00:00:00:00"
			;;
		adhoc)
			json_get_var nosbeacon sw_merge 1
			;;
	esac
	[ "$nosbeacon" = 1 ] || nosbeacon=""
	[ -n "${DEBUG}" ] && echo wlanconfig "$ifname" create wlandev "$phy" wlanmode "$wlanmode" ${wlanaddr:+wlanaddr "$wlanaddr"} ${nosbeacon:+nosbeacon}
	ifname=$(/usr/sbin/wlanconfig "$ifname" create wlandev "$phy" wlanmode "$wlanmode" ${wlanaddr:+wlanaddr "$wlanaddr"} ${nosbeacon:+nosbeacon})
	[ $? -ne 0 ] && {
		echo "qcawifi: Failed to set up $mode vif $ifname"
		return
	}
#	config_set "$vif" ifname "$ifname"
	pureg=0
	case "$hwmode:$htmode" in
	# The parsing stops at the first match so we need to make sure
	# these are in the right orders (most generic at the end)
		*g:HT20) hwmode=11NGHT20;;
		*g:HT40)
			hwmode=11NGHT40MINUS
			[ "$channel" -lt "7" ] && hwmode=11NGHT40PLUS
		;;
		*g:*) hwmode=11G;;
		*a:HT20) hwmode=11NAHT20;;
		*a:HT40)
			hwmode=11NAHT40MINUS
			[ "$(( ($channel / 4) % 2 ))" == "1" ] && hwmode=11NAHT40PLUS
		;;
		*a:VHT20) hwmode=11ACVHT20;;
		*a:VHT40) hwmode=11ACVHT40;;
		*a:VHT80) hwmode=11ACVHT80;;
		*a:VHT160) hwmode=11ACVHT160;;
		*a:*) hwmode=11A;;
		*) hwmode=AUTO;;
	esac
	iwpriv "$ifname" mode "$hwmode"
	[ $pureg -gt 0 ] && iwpriv "$ifname" pureg "$pureg"
	json_get_var cfreq2 cfreq2
	[ -n "$cfreq2" -a "$htmode" = "HT80_80" ] && iwpriv "$ifname" cfreq2 "$cfreq2"
	apply_config_to_if "$ifname" puren
	iwconfig "$ifname" channel "$channel" #>/dev/null 2>/dev/null
	json_get_vars hidden
	set_default hidden 0
	iwpriv "$ifname" hide_ssid "$hidden"
	json_get_var shortgi shortgi 1
	[ -n "$shortgi" ] && iwpriv "$ifname" shortgi "${shortgi}"
	json_get_var noscan noscan $noscan
	[ -n "$noscan" ] && iwpriv "$ifname" disablecoext "${noscan}"
	apply_config_to_if "$ifname" chwidth
	json_get_var wds wds
	case "$wds" in
		1|on|enabled) wds=1;;
		*) wds=0;;
	esac
	iwpriv "$ifname" wds "$wds" >/dev/null 2>&1
	apply_config_to_if "$ifname" TxBFCTL bintval countryie
	case "$enc" in
		none)
			# If we're in open mode and want to use WPS, we
			# must start hostapd
			json_get_var wps_pbc wps_pbc 0
			json_get_var config_methods wps_config
			[ "$wps_pbc" -gt 0 ] && append config_methods push_button
			[ -n "$config_methods" ] && start_hostapd=1
		;;
		wep*)
			case "$enc" in
				*mixed*)  iwpriv "$ifname" authmode 4;;
				*shared*) iwpriv "$ifname" authmode 2;;
				*)        iwpriv "$ifname" authmode 1;;
			esac
			for idx in 1 2 3 4; do
				json_get_var key "key${idx}"
				iwconfig "$ifname" enc "[$idx]" "${key:-off}"
			done
			json_get_var key key
			key="${key:-1}"
			case "$key" in
				[1234]) iwconfig "$ifname" enc "[$key]";;
				*) iwconfig "$ifname" enc "$key";;
			esac
		;;
		mixed*|psk*|wpa*|8021x)
			start_hostapd=1
			json_get_var key key
		;;
		wapi*)
			start_wapid=1
			json_get_var key key
		;;
	esac
	case "$mode" in
		sta|adhoc)
			json_get_var addr bssid
			[ -z "$addr" ] || {
				iwconfig "$ifname" ap "$addr"
			}
		;;
	esac
	json_get_vars uapsd
	set_default uapsd 0
	iwpriv "$ifname" uapsd "$uapsd"
	json_get_var mcast_rate mcast_rate
	[ -n "$mcast_rate" ] && iwpriv "$ifname" mcast_rate "${mcast_rate%%.*}"
	apply_config_to_if "$ifname" powersave ant_ps_on ps_timeout mcastenhance metimer metimeout medropmcast me_adddeny vap_ind extap scanband periodicScan
	json_get_var frag frag
	[ -n "$frag" ] && iwconfig "$ifname" frag "${frag%%.*}"
	json_get_var rts rts
	[ -n "$rts" ] && iwconfig "$ifname" rts "${rts%%.*}"
	apply_config_to_if "$ifname" cwmin cwmax aifs txoplimit noackpolicy
	json_get_vars wmm
	set_default wmm 0
	[ -n "$wmm" ] && iwpriv "$ifname" wmm "$wmm"
	json_get_vars doth
	set_default doth 0
	[ -n "$doth" ] && iwpriv "$ifname" doth "$doth"
	apply_config_to_if "$ifname" doth_chanswitch quiet mfptest dtim_period noedgech ps_on_time inact wnm ampdu amsdu maxampdu vhtmaxampdu setaddbaoper addbaresp addba delba
	json_get_var stafwd stafwd 0
	[ -n "$stafwd" ] && iwpriv "$ifname" stafwd "$stafwd"
	json_get_values maclist maclist
	[ -n "$maclist" ] && {
		# flush MAC list
		iwpriv "$ifname" maccmd 3
		for mac in $maclist; do
			iwpriv "$ifname" addmac "$mac"
		done
	}
	json_get_var macfilter macfilter
	case "$macfilter" in
		allow)
			iwpriv "$ifname" maccmd 1
		;;
		deny)
			iwpriv "$ifname" maccmd 2
		;;
		*)
			# default deny policy if mac list exists
			[ -n "$maclist" ] && iwpriv "$ifname" maccmd 2
		;;
	esac
	apply_config_to_if "$ifname" nss vht_mcsmap chwidth chbwmode ldpc rx_stbc tx_stbc cca_thresh set11NRetries chanbw maxsta sko_max_xretries extprotmode extprotspac
	apply_config_to_if "$ifname" cwmenable protmode enablertscts txcorrection rxcorrection
	json_get_var ssid ssid
	[ -n "$ssid" ] && {
		iwconfig "$ifname" essid on
		iwconfig "$ifname" essid "$ssid"
	}
	apply_config_to_if "$ifname" txqueuelen
# FIXME:
#	[ -n "$mtu" ] && {
#		config_set "$vif" mtu $mtu
#		ifconfig "$ifname" mtu $mtu
#	}
	apply_config_to_if "$ifname" tdls set_tdls_rmac tdls_qosnull tdls_uapsd tdls_set_rcpi tdls_set_rcpi_hi tdls_set_rcpi_lo tdls_set_rcpi_margin \
		tdls_dtoken do_tdls_dc_req tdls_auto tdls_off_timeout tdls_tdb_timeout tdls_weak_timeout tdls_margin tdls_rssi_ub tdls_rssi_lb
	json_get_var tdls_path_sel tdls_path_sel
	[ -n "$tdls_path_sel" ] && iwpriv "$ifname" tdls_pathSel "$tdls_path_sel"
	json_get_var tdls_rssi_offset tdls_rssi_offset
	[ -n "$tdls_rssi_offset" ] && iwpriv "$ifname" tdls_rssi_o "$tdls_rssi_offset"
	json_get_var tdls_path_sel_period tdls_path_sel_period
	[ -n "$tdls_path_sel_period" ] && iwpriv "$ifname" tdls_pathSel_p "$tdls_path_sel_period"
	apply_config_to_if "$ifname" tdlsmacaddr1 tdlsmacaddr2 tdlsaction tdlsoffchan tdlsswitchtime tdlstimeout tdlsecchnoffst tdlsoffchnmode blockdfschan dbgLVL \
		acsmindwell acsmaxdwell acsreport ch_hop_en ch_long_dur ch_nhop_dur ch_cntwn_dur ch_noise_th ch_cnt_th scanchevent send_add_ies ext_ifu_acs rrm rrmslwin \
		rrmsstats rrmdbg acparams setwmmparams qbssload proxyarp dgaf_disable setibssdfsparam startibssrssimon setibssrssihyst noIBSSCreate setibssrssiclass \
		offchan_tx_test
	json_get_values vow_dbg_cfg_values vow_dbg_cfg
	[ -n "$vow_dbg_cfg_values" ] && {
		for value in $vow_dbg_cfg_values
		do
			iwpriv "$ifname" vow_dbg_cfg $value
		done
	}
	apply_config_to_if "$ifname" vow_dbg
	json_get_values set_max_rate_values set_max_rate
	[ -n "$set_max_rate_values" ] && {
		for value in $set_max_rate_values
		do
			wlanconfig "$ifname" set_max_rate $value
		done
	}
	apply_config_to_if "$ifname" implicitbf vhtsubfee vhtmubfee vhtsubfer vhtmubfer vhtstscap vhtsounddim encap_type decap_type rawsim_txagr clr_rawsim_stats rawsim_debug \
		set_monrxfilter rxfilterinvalid athnewind osen
#!!!
	json_get_vars isolate
	set_default isolate 0
	echo "qcawifi:Bring $ifname up now."
	ifconfig "$ifname" up
	case "$mode" in
		ap|wrap|ap_monitor|ap_smart_monitor|mesh)

			iwpriv "$ifname" ap_bridge "$((isolated^1))"
			json_get_var l2tif l2tif
			[ -n "$l2tif" ] && iwpriv "$ifname" l2tif "$l2tif"
			if [ -n "$start_wapid" ]; then
#				wapid_setup_vif "$vif" || {
#					echo "enable_qcawifi($device): Failed to set up wapid for interface $ifname" >&2
					ifconfig "$ifname" down
					wlanconfig "$ifname" destroy
					return
#				}
			fi
			if [ -n "$start_hostapd" ] ; then
				qcawifi_hostapd_setup_vif "$ifname" || {
					echo "enable_qcawifi($device): Failed to set up hostapd for interface $ifname"
					# make sure this wifi interface won't accidentally stay open without encryption
					ifconfig "$ifname" down
					wlanconfig "$ifname" destroy
					return
				}
			fi
		;;
		wds|sta)
			qcawifi_wpa_supplicant_setup_vif "$ifname" || {
				echo "enable_qcawifi($device): Failed to set up wpa_supplicant for interface $ifname" >&2
				ifconfig "$ifname" down
				wlanconfig "$ifname" destroy
				return
			}
		;;
		adhoc)
			qcawifi_wpa_supplicant_setup_vif "$ifname" || {
				echo "enable_qcawifi($device): Failed to set up wpa"
				ifconfig "$ifname" down
				wlanconfig "$ifname" destroy
				return
			}
	esac
	wireless_add_vif "$name" "$ifname"
	json_get_vars bridge
	[ -z `brctl show | grep $ifname` ] && [ ! -z $bridge ] && {
		echo "qcawifi:Manually bridge interface $ifname into $bridge"
		brctl addif $bridge $ifname 
	}
	apply_config_to_if "$ifname" set11NRates
	# 256 QAM capability needs to be parsed first, since
	# vhtmcs enables/disable rate indices 8, 9 for 2G
	# only if vht_11ng is set or not
	apply_config_to_if "$ifname" vht_11ng vhtmcs
	#support nawds
#	json_get_var nawds_mode nawds_mode
#	[ -n "$nawds_mode" ] && wlanconfig "$ifname" nawds mode "${nawds_mode}"
#
#	handle_nawds() {
#		local value="$1"
#		[ -n "$value" ] && wlanconfig "$ifname" nawds add-repeater $value
#	}
#	config_list_foreach "$vif" nawds_add_repeater handle_nawds
#
#	handle_hmwds() {
#		local value="$1"
#		[ -n "$value" ] && wlanconfig "$ifname" hmwds add_addr $value
#	}
#	config_list_foreach "$vif" hmwds_add_addr handle_hmwds
#
#	json_get_var nawds_override nawds_override
#	[ -n "$nawds_override" ] && wlanconfig "$ifname" nawds override "${nawds_override}"
#
#	json_get_var nawds_defcaps nawds_defcaps
#	[ -n "$nawds_defcaps" ] && wlanconfig "$ifname" nawds defcaps "${nawds_defcaps}"
#
#	handle_hmmc_add() {
#		local value="$1"
#		[ -n "$value" ] && wlanconfig "$ifname" hmmc add $value
#	}
#	config_list_foreach "$vif" hmmc_add handle_hmmc_add
	# TXPower settings only work if device is up already
	# while atheros hardware theoretically is capable of per-vif (even per-packet) txpower
	# adjustment it does not work with the current atheros hal/madwifi driver
	json_get_var vif_txpower txpower
	# use vif_txpower (from wifi-iface) instead of txpower (from wifi-device) if
	# the latter doesn't exist
	txpower="${txpower:-$vif_txpower}"
	[ -z "$txpower" ] || iwconfig "$ifname" txpower "${txpower%%.*}"
	json_select ..
	return 0;
}

drv_qcawifi_setup() {
	# 这里使用qca_device变量; device是全局变量;避免重名enable_smp_affinity_wifi(); 避免死锁
	qca_device="$1"
	lock /var/run/qcawifi-$qca_device.lock

	#FIXME:
	# Workaround for qcawifi oops if system memory less
	# maybe we can try set /proc/sys/vm/vfs_cache_pressure ?
	echo "qcawifi: drop cache before setup device $qca_device"
	sync && sysctl -w vm.drop_caches=3
	
	echo "qcawifi:setup device $qca_device"
	check_qcawifi_phy "$qca_device" || {
		wireless_set_retry 0
		lock -u /var/run/qcawifi-$qca_device.lock
		return 1
	}

	json_select config
	json_get_var macaddr macaddr "$(cat /sys/class/net/${qca_device}/address)"
	json_select ..
	echo "qcawifi:set data $qca_device"
	wireless_set_data phy="$qca_device"
	local phy="$qca_device"
	json_select config
	# If the country parameter is number (either hex or decimal), we
	# assume it's a regulatory domain - i.e. we use iwpriv setCountryID.
	# Else we assume it's a country code - i.e. we use iwpriv setCountry.
	json_get_var country country
	if [ `expr "$country" : '[0-9].*'` -ne 0 ]; then
		iwpriv "$phy" setCountryID "$country"
	elif [ -n "$country" ]; then
		iwpriv "$phy" setCountry "$country"
	fi
	json_get_var channel channel
	json_get_var txpower txpower
	[ auto = "$channel" ] && channel=0
	json_get_var antdiv diversity
	json_get_var antrx rxantenna
	json_get_var anttx txantenna
	json_get_var softled softled
	json_get_var antenna antenna
	json_get_var distance distance
	[ -n "$antdiv" ] && echo "antdiv option not supported on this driver"
	[ -n "$antrx" ] && echo "antrx option not supported on this driver"
	[ -n "$anttx" ] && echo "anttx option not supported on this driver"
	[ -n "$softled" ] && echo "softled option not supported on this driver"
	[ -n "$antenna" ] && echo "antenna option not supported on this driver"
	[ -n "$distance" ] && echo "distance option not supported on this driver"
	# Advanced QCA wifi per-radio parameters configuration
	apply_config_to_if "$phy" txchainmask rxchainmask AMPDU ampdudensity AMSDU AMPDULim AMPDUFrames AMPDURxBsize
	json_get_var bcnburst bcnburst 0
	[ "$bcnburst" -gt 0 ] && iwpriv "$phy" set_bcnburst "$bcnburst"
	apply_config_to_if "$phy" set_smart_antenna current_ant default_ant ant_retrain retrain_interval retrain_drop \
		ant_train ant_trainmode ant_traintype ant_pktlen ant_numpkts ant_numitr
	json_get_var ant_train_thres ant_train_thres
	[ -n "$ant_train_thres" ] && iwpriv "$phy" train_threshold "$ant_train_thres"
	json_get_var ant_train_min_thres ant_train_min_thres
	[ -n "$ant_train_min_thres" ] && iwpriv "$phy" train_threshold "$ant_train_min_thres"
	apply_config_to_if "$phy" ant_traffic_timer dcs_enable dcs_coch_int dcs_errth dcs_phyerrth dcs_usermaxc dcs_debug
	json_get_var set_ch_144 set_ch_144
	[ -n "$set_ch_144" ] && iwpriv "$phy" setCH144 "$set_ch_144"
	json_get_var eppovrd_ch_144 eppovrd_ch_144
	[ -n "$eppovrd_ch_144" ] && iwpriv "$phy" setCH144EppOvrd "$eppovrd_ch_144"
	apply_config_to_if "$phy" ani_enable acs_bkscanen acs_scanintvl acs_rssivar acs_chloadvar acs_lmtobss acs_ctrlflags acs_dbgtrace dscp_ovride reset_dscp_map dscp_tid_map
	json_get_var igmp_dscp_ovride igmp_dscp_ovride
	[ -n "$igmp_dscp_ovride" ] && iwpriv "$phy" sIgmpDscpOvrid "$igmp_dscp_ovride"
	json_get_var igmp_dscp_tid_map igmp_dscp_tid_map
	[ -n "$igmp_dscp_tid_map" ] && iwpriv "$phy" sIgmpDscpTidMap "$igmp_dscp_tid_map"
	json_get_var hmmc_dscp_ovride hmmc_dscp_ovride
	[ -n "$hmmc_dscp_ovride" ] && iwpriv "$phy" sHmmcDscpOvrid "$hmmc_dscp_ovride"
	json_get_var hmmc_dscp_tid_map hmmc_dscp_tid_map
	[ -n "$hmmc_dscp_tid_map" ] && iwpriv "$phy" sHmmcDscpTidMap "$hmmc_dscp_tid_map"
	json_get_var blk_report_fld blk_report_fld
	[ -n "$blk_report_fld" ] && iwpriv "$phy" setBlkReportFld "$blk_report_fld"
	json_get_var drop_sta_query drop_sta_query
	[ -n "$drop_sta_query" ] && iwpriv "$phy" setDropSTAQuery "$drop_sta_query"
	apply_config_to_if "$phy" burst burst_dur TXPowLim2G TXPowLim5G enable_ol_stats rst_tso_stats rst_lro_stats rst_sg_stats \
		set_fw_recovery allowpromisc set_sa_param aldstats macaddr promisc
	json_get_values aggr_burst aggr_burst

	[ -n "$aggr_burst" ] && {
		for value in $aggr_burst
		do
			iwpriv "$phy" aggr_burst $value
		done
	}
	json_get_var block_interbss block_interbss
	[ -n "$block_interbss" ] && iwpriv "$phy" block_interbss "$block_interbss"
	json_get_var set_pmf set_pmf
	[ -n "$set_pmf" ] && iwpriv "$phy" set_pmf "${set_pmf}"
	json_get_var txbf_snd_int txbf_snd_int 100
	[ -n "$txbf_snd_int" ] && iwpriv "$phy" txbf_snd_int "$txbf_snd_int"
	json_get_var mcast_echo mcast_echo
	[ -n "$mcast_echo" ] && iwpriv "$phy" mcast_echo "${mcast_echo}"
	json_get_var obss_rssi_th obss_rssi_th 35
	[ -n "$obss_rssi_th" ] && iwpriv "$phy" obss_rssi_th "${obss_rssi_th}"
	json_get_var obss_rx_rssi_th obss_rx_rssi_th 35
	[ -n "$obss_rx_rssi_th" ] && iwpriv "$phy" obss_rx_rssi_th "${obss_rx_rssi_th}"
	json_get_var acs_txpwr_opt acs_txpwr_opt
	[ -n "$acs_txpwr_opt" ] && iwpriv "$phy" acs_tcpwr_opt "${acs_txpwr_opt}"
	json_get_var obss_long_slot obss_long_slot
	[ -n "$obss_long_slot" ] && iwpriv "$phy" obss_long_slot "${obss_long_slot}"
	json_get_var staDFSEnable staDFSEnable
	[ -n "$staDFSEnable" ] && iwpriv "$phy" staDFSEnable "${staDFSEnable}"
#options for vifs from device
	json_get_var nosbeacon nosbeacon
	json_get_var qwrap_enable qwrap_enable 0
	json_get_var noscan noscan
	json_get_var hwmode hwmode auto
	json_get_var htmode htmode auto
	json_select ..

#count vifs
	cntvar=0
	qcawifi_vif_count() {
		let cntvar+=1
	}
	for_each_interface "sta" qcawifi_vif_count
	local sta_cnt=$cntvar
	cntvar=0
	for_each_interface "ap" qcawifi_vif_count
	local ap_cnt=$cntvar
	echo "qcawifi:Found $sta_cnt STA interface(s) and $ap_cnt AP interface(s)."
	nosbeacon=""
	[ $sta_cnt -ge 1 ] && [ $ap_cnt -ge 1 ] && {
		echo "qcawifi:Disable beacon timers for STA."
		nosbeacon=1
	}
	radioidx=${qca_device#wifi}
	ifidx=0
#setup vifs
	for_each_interface "adhoc sta ap monitor wrap ap_monitor ap_smart_monitor mesh" qcawifi_vif_setup
	if [ -f "/lib/update_smp_affinity.sh" ]; then
		. /lib/update_smp_affinity.sh
		config_load "wireless"
		# device是全局变量; 避免重名enable_smp_affinity_wifi()
		config_foreach enable_smp_affinity_wifi wifi-device
	fi
#	json_select config
#	json_select ..
	[ -f /etc/config/qcawifi ] && {
		local icm_enable qwrap_enable
	
		config_load "qcawifi"
	
		config_get_bool icm_enable icm enable 0
		[ ${icm_enable} -gt 0 ] && \
				eval "type icm_setup" >/dev/null 2>&1 && {
			icm_setup
		}
		config_get_bool wpc_enable wpc enable 0
		[ ${wpc_enable} -gt 0 ] && \
				eval "type wpc_setup" >/dev/null 2>&1 && {
			wpc_setup
		}
	}
	eval "type qwrap_setup" >/dev/null 2>&1 && qwrap_setup

	# The init script will check whether lbd is actually
	# enabled
	[ -f /etc/init.d/lbd ] && /etc/init.d/lbd restart
	[ -f /etc/init.d/ssid_steering ] &&  /etc/init.d/ssid_steering restart

	wireless_set_up
	lock -u /var/run/qcawifi-$qca_device.lock
}

drv_qcawifi_teardown() {
	json_select data
	json_get_var phy phy
	json_select ..
	eval "type qwrap_teardown" >/dev/null 2>&1 && qwrap_teardown
	eval "type icm_teardown" >/dev/null 2>&1 && icm_teardown
	eval "type wpc_teardown" >/dev/null 2>&1 && wpc_teardown
	[ ! -f /etc/init.d/lbd ] || /etc/init.d/lbd stop
	[ ! -f /etc/init.d/ssid_steering ] || /etc/init.d/ssid_steering stop
	local device="$1"
	local parent
	check_qcawifi_phy "$device" || return 1
	include /lib/network
	cd /sys/class/net
	for dev in *; do
		[ -f /sys/class/net/${dev}/parent ] && { \
			local parent=$(cat /sys/class/net/${dev}/parent)
			[ -n "$parent" -a "$parent" = "$device" ] && { \
				[ -f "/var/run/wifi-${dev}.pid" ] &&
					kill "$(cat "/var/run/wifi-${dev}.pid")"
				[ -f "/var/run/hostapd-${dev}.lock" ] && { \
					wpa_cli -g /var/run/hostapd/global raw REMOVE ${dev}
					rm /var/run/hostapd-${dev}.lock
				}
				[ -f "/var/run/wpa_supplicant-${dev}.lock" ] && { \
					wpa_cli -g /var/run/wpa_supplicantglobal interface_remove ${dev}
					rm /var/run/wpa_supplicant-${dev}.lock
				}
				ifconfig "$dev" down
				unbridge "$dev"
				wlanconfig "$dev" destroy
			}
			[ -f /var/run/hostapd_cred_${device}.bin ] && { \
				rm /var/run/hostapd_cred_${device}.bin
			}
		}
	done
	return 0
}

drv_qcawifi_cleanup() {
	cd /sys/class/net
	for wdev in wifi*; do
		drv_qcawifi_teardown $wdev
	done
}
add_driver qcawifi
