#!/bin/sh
#
# Copyright (c) 2014, The Linux Foundation. All rights reserved.
# Copyright (c) 2016, GuoGuo<gch981213@gmail.com>
#
#  Permission to use, copy, modify, and/or distribute this software for any
#  purpose with or without fee is hereby granted, provided that the above
#  copyright notice and this permission notice appear in all copies.
#
#  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
#  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
#  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
#  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
#  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
#  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
#  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

append DRIVERS "qcawifi"

wlanconfig() {
	[ -n "${DEBUG}" ] && echo wlanconfig "$@"
	/usr/sbin/wlanconfig "$@"
}

iwconfig() {
	[ -n "${DEBUG}" ] && echo iwconfig "$@"
	/usr/sbin/iwconfig "$@"
}

iwpriv() {
	[ -n "${DEBUG}" ] && echo iwpriv "$@"
	/usr/sbin/iwpriv "$@"
}

find_qcawifi_phy() {
	local device="$1"

	local macaddr="$(config_get "$device" macaddr | tr 'A-Z' 'a-z')"
	config_get phy "$device" phy
	[ -z "$phy" -a -n "$macaddr" ] && {
		cd /sys/class/net
		for phy in $(ls -d wifi* 2>&-); do
			[ "$macaddr" = "$(cat /sys/class/net/${phy}/address)" ] || continue
			config_set "$device" phy "$phy"
			break
		done
		config_get phy "$device" phy
	}
	[ -n "$phy" -a -d "/sys/class/net/$phy" ] || {
		echo "phy for wifi device $1 not found"
		return 1
	}
	[ -z "$macaddr" ] && {
		config_set "$device" macaddr "$(cat /sys/class/net/${phy}/address)"
	}
	return 0
}

check_qcawifi_device() {
	[ ${1%[0-9]} = "wifi" ] && config_set "$1" phy "$1"
	config_get phy "$1" phy
	[ -z "$phy" ] && {
		find_qcawifi_phy "$1" >/dev/null || return 1
		config_get phy "$1" phy
	}
	[ "$phy" = "$dev" ] && found=1
}


detect_qcawifi() {
	devidx=0
	config_load wireless
	while :; do
		config_get type "wifi$devidx" type
		[ -n "$type" ] || break
		devidx=$(($devidx + 1))
	done
	cd /sys/class/net
	[ -d wifi0 ] || return
	for dev in $(ls -d wifi* 2>&-); do
		found=0
		config_foreach check_qcawifi_device wifi-device
		[ "$found" -gt 0 ] && continue

		hwcaps=$(cat ${dev}/hwcaps)
		case "${hwcaps}" in
			*11bgn |\
			*11abgn)
				mode_hw=11g
				mode_ht=HT20
				noscan=0
				#2.4G no pad
				#ssid_pad="2.4G"
				ssid_pad=""
			;;
			*11an)
				mode_hw=11a
				mode_ht=HT40
				ssid_pad="_5G"
			;;
			*11an/ac |\
			*11abgn/ac)
				mode_hw=11a
				if grep -q VHT80_80 ${dev}/hwmodes
				then
					mode_ht=VHT160
				else
					mode_ht=VHT80
				fi
				ssid_pad="_5G"
				noscan=1
			;;
		esac
		
		#uci set wireless.default_wifi${devidx}.ssid=PandoraBox-${ssid_pad}-$(cat /sys/class/net/${dev}/address | awk -F ":" '{print $4""$5""$6 }'| tr a-z A-Z)
		
		uci -q batch <<-EOF
			set wireless.wifi${devidx}=wifi-device
			set wireless.wifi${devidx}.type=qcawifi
			set wireless.wifi${devidx}.channel=auto
			set wireless.wifi${devidx}.country='CN'
			set wireless.wifi${devidx}.hwmode=${mode_hw}
			set wireless.wifi${devidx}.htmode=${mode_ht}
			set wireless.wifi${devidx}.noscan=$(noscan)
			set wireless.default_wifi${devidx}=wifi-iface
			set wireless.default_wifi${devidx}.device=wifi$devidx
			set wireless.default_wifi${devidx}.network=lan
			set wireless.default_wifi${devidx}.mode=ap
			set wireless.default_wifi${devidx}.ssid=$(cat /proc/sys/kernel/hostname)${ssid_pad}
			set wireless.default_wifi${devidx}.encryption=none
			set wireless.default_wifi${devidx}.wmm=1
EOF
	uci -q commit wireless
	devidx=$(($devidx + 1))
	done
}
